from docx.shared import Parented
from docx.text.run import Run


class Revision(Parented):

    def __init__(self, rev, parent):
        super(Revision, self).__init__(parent)
        self._rev = self._element = self.element = rev

    @property
    def runs(self):
        return [Run(r, self) for r in self._rev.r_lst]

    @property
    def text(self):
        text = ''
        for run in self.runs:
            text += run.text
            text += run.del_text
        return text

    @property
    def tagroot(self):
        nspfx, tagroot = self.element._nsptag.split(':')
        return tagroot


class Deltext(Parented):

    def __init__(self, delt, parent):
        super(Deltext, self).__init__(parent)
        self._delt = self._element = self.element = delt




import docx


def get_ins_del(p):
    from pymdocx.common.utils import print_xml_node
    ins_list = get_ins_in_paragraph(p)
    for ins in ins_list:
        print_xml_node(ins)
    del_list = get_del_in_paragraph(p)
    for del_ in del_list:
        print_xml_node(del_)


def get_revision(doc_file_path):
    document = docx.Document(docx=doc_file_path)
    for p in document.paragraphs:
        get_ins_del(p)


def has_revision_ins(p):
    return 1 if get_ins_in_paragraph(p) else 0


def has_revision_del(p):
    return 1 if get_del_in_paragraph(p) else 0


def has_revision(p):
    return has_revision_del(p) or has_revision_ins(p)


def get_ins_in_paragraph(p):
    from pymdocx.common.utils import get_label_in_paragraph
    # 获取新增修订
    return get_label_in_paragraph(p, 'ins')


def get_del_in_paragraph(p):
    from pymdocx.common.utils import get_label_in_paragraph
    # 获取删除修订
    return get_label_in_paragraph(p, 'del')


def parse_p_revision(p_obj):
    rev_list = []
    for r in p_obj.revisions:
        text = '增加：' + r.text if r.tagroot == 'ins' else '删除：' + r.text
        rev_list.append({
            'tagroot': r.tagroot,
            'text': text,
            'dtime': r.element.date,
            'author': r.element.author,
            'initials': '',
            'id': r.element._id,
        })
    return rev_list


def add_revision_2_p_end(p1, p2, comment_part_obj):
    rev_list = parse_p_revision(p2)
    for r in rev_list:
        p1.add_comment(r['text'],
                       author=r['author'],
                       dtime=r['dtime'],
                       comment_part=comment_part_obj,
                       rangeStart=len(p1._p.getchildren()),
                       rangeEnd=len(p1._p.getchildren()))


def remove_revision(p):
    for r in p.revisions:
        pp = r._rev.getparent()
        if r.tagroot == 'ins':
            pp.remove(r._rev)
        elif r.tagroot == 'del':
            new_run = p.add_run(r.text, '4')
            pp.replace(r._rev, new_run._r)


import os

from pymdocx.common.utils import get_doc, print_xml_node
from pymdocx.doc.table import merge_table_comment_revision

DIR_PATH = './../data/test_table'
OUTPUT_PATH = './../data/output'


def test_merge_table_comment_revision():
    doc_o_path = os.path.join(DIR_PATH, 't_base.docx')
    doc_a_path = os.path.join(DIR_PATH, 't_base_1.docx')
    doc_b_path = os.path.join(DIR_PATH, 't_base_2.docx')
    output_file_path = os.path.join(DIR_PATH, "t_base_new.docx")

    doc_o = get_doc(doc_o_path)
    doc_a = get_doc(doc_a_path)
    doc_b = get_doc(doc_b_path)
    merge_table_comment_revision(doc_o, [doc_a, doc_b])
    doc_o.save(output_file_path)


def test_table_xml():
    doc_a_path = os.path.join(DIR_PATH, 'a.docx')
    doc_a = get_doc(doc_a_path)
    table = doc_a.tables[0]
    print_xml_node(table._element)


import docx
from lxml import etree
import numpy as np

from pymdocx.common.comment import has_comment
from pymdocx.common.constants import WORD_NAMESPACE
from pymdocx.common.revision import has_revision


def get_label_in_paragraph(p, label_name):
    full_label_name = 'w:' + label_name
    xml = p._p.xml
    if full_label_name in xml:
        tree = etree.fromstring(xml)
        ins = tree.xpath('//' + full_label_name, namespaces=WORD_NAMESPACE)
    else:
        ins = []
    return ins


def print_xml_node(lxml_obj):
    print(etree.tostring(lxml_obj, pretty_print=True, encoding='utf-8').decode("utf-8"))


def print_doc(doc_file_path):
    document = docx.Document(docx=doc_file_path)
    print(document._body._element.xml)


def get_doc(doc_file_path):
    document = docx.Document(docx=doc_file_path)
    return document


def get_element_comment_revision_matrix(element):
    comment_list = []
    revision_list = []
    for p in element.paragraphs:
        comment_list.append(has_comment(p))
        revision_list.append(has_revision(p))
    return np.array([comment_list, revision_list])


def add_p_next(p1, p2):
    p1._p.addnext(p2._p)


def _get_actual_p_index(has_add_mapping, doc_index, p_index):
    if doc_index in has_add_mapping.keys():
        has_add_mapping[doc_index] += 1
    else:
        has_add_mapping[doc_index] = 0
    return p_index - has_add_mapping[doc_index]


import numpy as np

from pymdocx.common.comment import add_p_comment_next, add_comment_2_p_end, has_comment
from pymdocx.common.revision import add_revision_2_p_end, remove_revision, has_revision
from pymdocx.common.utils import get_element_comment_revision_matrix, \
    _get_actual_p_index


def get_merge_res(m_list):
    # 返回一个字典，每个被批注和修订的段落的索引，以及对应的文档索引
    # key: p index
    # value: 文件 index
    #
    # {0: [0], 1: [1, 0], 2: [0], 8: [0], 10: [0]}
    m_merge = np.array(m_list)
    print('m_merge:')
    print(m_merge)

    print('merge doc modify')
    m_modify = np.max(m_merge, axis=1)
    print(m_modify)

    print('修改')
    m_arg = np.argwhere(m_modify > 0)
    print(m_arg)

    merge_arg_dict = {}
    for doc_index, p_index in m_arg:
        if merge_arg_dict.get(p_index):
            merge_arg_dict.get(p_index).append(doc_index)
        else:
            merge_arg_dict[p_index] = [doc_index]
    return merge_arg_dict


def p_bold_italic(p):
    for r in p.runs:
        r.bold = True
        r.italic = True


def merge_paragraph_comment_revision_stack(doc_base_obj, doc_list):
    m_list = [get_element_comment_revision_matrix(doc) for doc in doc_list]
    merge_arg_dict = get_merge_res(m_list)

    has_add_mapping = {}
    has_add_p_count = 0
    remove_p_list = []
    for p_index, doc_index_list in merge_arg_dict.items():
        last_p = doc_base_obj.paragraphs[p_index + has_add_p_count]
        remove_p_list.append(last_p)
        for i, doc_index in enumerate(doc_index_list):
            actual_p_index = _get_actual_p_index(has_add_mapping, doc_index, p_index)
            target_p = doc_list[doc_index].paragraphs[actual_p_index]
            # 添加段落后，target_p所在文档结构发生变化
            add_p_comment_next(last_p, target_p, doc_base_obj.comments_part.element)
            last_p = target_p
            if i > 0:
                p_bold_italic(last_p)
            has_add_p_count += 1
    [rp.delete() for rp in remove_p_list]


def merge_paragraph_comment_revision_end(doc_base_obj, doc_list):
    m_list = [get_element_comment_revision_matrix(doc) for doc in doc_list]
    merge_arg_dict = get_merge_res(m_list)

    has_add_mapping = {}
    has_add_p_count = 0
    remove_p_list = []
    for p_index, doc_index_list in merge_arg_dict.items():
        last_p = doc_base_obj.paragraphs[p_index + has_add_p_count]
        remove_p_list.append(last_p)
        for i, doc_index in enumerate(doc_index_list):
            merge_doc_paragraphs = doc_list[doc_index].paragraphs
            has_add_p_count, last_p = _merge_p(i, has_add_mapping, doc_index, p_index, merge_doc_paragraphs, last_p,
                                               doc_base_obj.comments_part.element, has_add_p_count, remove_p_list)

    [rp.delete() for rp in remove_p_list]


def _merge_p(i, has_add_mapping, doc_index, p_index, merge_doc_paragraphs,
             last_p, comments_part_obj, has_add_p_count, remove_p_list):
    if i == 0:
        target_p = merge_doc_paragraphs[_get_actual_p_index(has_add_mapping, doc_index, p_index)]
        if has_comment(target_p):
            add_p_comment_next(last_p, target_p, comments_part_obj)
            last_p = target_p
            has_add_p_count += 1
        else:
            has_add_mapping[doc_index] -= 1
            remove_p_list.remove(last_p)
        if has_revision(target_p):
            add_revision_2_p_end(last_p, target_p, comments_part_obj)
            remove_revision(last_p)
    else:
        add_comment_2_p_end(last_p, merge_doc_paragraphs[p_index], comments_part_obj)
        add_revision_2_p_end(last_p, merge_doc_paragraphs[p_index], comments_part_obj)
    return has_add_p_count, last_p


def parse_paragraph_differences(doc_base_obj, doc_list):
    m_num = len(doc_list)
    merge_dict = {}
    dm_cursor = [-1 for _ in range(m_num)]
    for dbpi, dbp in enumerate(doc_base_obj.paragraphs):
        merge_dict[dbpi] = {"correspond_p": [None for _ in range(m_num)], "new_p": [[] for _ in range(m_num)], 'correspond_p_m': [0 for _ in range(m_num)]}
        for dmi, d in enumerate(doc_list):
            next_p = dm_cursor[dmi] + 1
            for dmpi, dmp in enumerate(d.paragraphs[next_p:], start=next_p):
                dm_cursor[dmi] = dmpi
                if dmp.origin_text.strip() == dbp.text.strip():
                    merge_dict[dbpi]['correspond_p'][dmi] = dmpi
                    if has_comment(dmp) or has_revision(dmp):
                        merge_dict[dbpi]['correspond_p_m'][dmi] = 1
                    break
                else:
                    merge_dict[dbpi]['new_p'][dmi].append(dmpi)
    return merge_dict


merge_paragraph_comment_revision = merge_paragraph_comment_revision_stack